/*
 * Copyright (c) 2023-2024 VMware Carbon Black, all rights reserved
 */

package consumer

import (
	"time"

	"github.com/rs/zerolog"
	"github.com/rs/zerolog/log"

	"github.com/wavefronthq/go-metrics-wavefront/reporting"

	"gitlab.bit9.local/psc/platform/ems/hello/last-contact-updater/internal/stats"
)

// KclMonitoringService sends kcl stats to wavefront
type KclMonitoringService struct {
	statsReporter reporting.WavefrontMetricsReporter
	log           *zerolog.Logger
}

// Init -- Initialize the service
//
//goland:noinspection GoMixedReceiverTypes
func (KclMonitoringService) Init(_, _, _ string) error { return nil }

// Start -- starts the service
//
//goland:noinspection GoMixedReceiverTypes
func (KclMonitoringService) Start() error { return nil }

// Shutdown -- Shutdown the service
//   - currently a no-op
//
//goland:noinspection GoMixedReceiverTypes
func (k *KclMonitoringService) Shutdown() {
}

// IncrRecordsProcessed -- update stats for processed records count
//
//goland:noinspection GoMixedReceiverTypes
func (k *KclMonitoringService) IncrRecordsProcessed(shard string, value int) {
	tags := make(map[string]string)
	tags["shard"] = shard
	tags["msc"] = stats.Mosaic2Tag
	stats.UpdateMeter(k.statsReporter, "kcl.process_records.count", tags, int64(value))
}

// IncrBytesProcessed -- update stats for processed records bytes
//
//goland:noinspection GoMixedReceiverTypes
func (k *KclMonitoringService) IncrBytesProcessed(shard string, bytes int64) {
	tags := make(map[string]string)
	tags["shard"] = shard
	tags["msc"] = stats.Mosaic2Tag
	stats.UpdateMeter(k.statsReporter, "kcl.process_records.bytes", tags, bytes)
}

// MillisBehindLatest -- update stats for consumer delay
//
//goland:noinspection GoMixedReceiverTypes
func (k *KclMonitoringService) MillisBehindLatest(shard string, millSeconds float64) {
	tags := make(map[string]string)
	tags["shard"] = shard
	tags["msc"] = stats.Mosaic2Tag
	stats.UpdateGauge(k.statsReporter, "kcl.shard.millis_behind_latest", tags, int64(millSeconds))
}

// LeaseGained -- update stats for lease gained
//
//goland:noinspection GoMixedReceiverTypes
func (KclMonitoringService) LeaseGained(shard string) {
	log.Warn().Msgf("Lease gained on shard with ID - %s", shard)
}

// LeaseLost -- update stats for lease lost
//
//goland:noinspection GoMixedReceiverTypes
func (k *KclMonitoringService) LeaseLost(shardID string) {
	log.Warn().Msgf("Lease lost on shard with ID - %s", shardID)
	log.Warn().Msgf("Unregistering shard level kcl metrics on shard - %s", shardID)
	tags := make(map[string]string)
	tags["shard"] = shardID
	tags["msc"] = stats.Mosaic2Tag
	stats.Unregister(k.statsReporter, "kcl.shard.millis_behind_latest", tags)
	stats.Unregister(k.statsReporter, "kcl.process_records.bytes", tags)
	stats.Unregister(k.statsReporter, "kcl.get_records.latency", tags)
	stats.Unregister(k.statsReporter, "kcl.process_records.latency", tags)
	stats.Unregister(k.statsReporter, "kcl.process_records.count", tags)
}

// LeaseRenewed  -- update stats for lease renewed
//   - currently only logs
//
//goland:noinspection GoMixedReceiverTypes
func (KclMonitoringService) LeaseRenewed(shard string) {
	log.Warn().Msgf("Lease renewed on shard with ID - %s", shard)
}

// RecordGetRecordsTime -- update stats for latency of get records
//
//goland:noinspection GoMixedReceiverTypes
func (k *KclMonitoringService) RecordGetRecordsTime(shard string, millis float64) {
	tags := make(map[string]string)
	tags["shard"] = shard
	tags["msc"] = stats.Mosaic2Tag
	stats.UpdateTimer(k.statsReporter, "kcl.get_records.latency", tags, time.Duration(millis)*time.Millisecond)
}

// RecordProcessRecordsTime -- update stats for latency of process records
//
//goland:noinspection GoMixedReceiverTypes
func (k *KclMonitoringService) RecordProcessRecordsTime(shard string, millis float64) {
	tags := make(map[string]string)
	tags["shard"] = shard
	tags["msc"] = stats.Mosaic2Tag
	stats.UpdateTimer(k.statsReporter, "kcl.process_records.latency", tags, time.Duration(millis)*time.Millisecond)
}
