/*
 * Copyright (c) 2023 VMware Carbon Black, all rights reserved
 */

package publisher

import (
	"fmt"
	"strconv"
	"time"

	"carbonblack.com/protocols/psc/devicelastcontacttime"
	"carbonblack.com/protocols/psc/ems/assetupdater"
	"github.com/tuvistavie/securerandom"
	"google.golang.org/protobuf/types/known/timestamppb"

	"gitlab.bit9.local/psc/platform/ems/hello/last-contact-updater/internal/config"

	"github.com/rs/zerolog"
	"github.com/wavefronthq/go-metrics-wavefront/reporting"
	"gitlab.bit9.local/psc/platform/orgkey-go/v3"
	elasticV5 "gopkg.in/olivere/elastic.v5"
	elasticV6 "gopkg.in/olivere/elastic.v6"

	"gitlab.bit9.local/psc/platform/ems/hello/last-contact-updater/internal/stats"
)

const (
	// LastContactFieldName lastContact field name in proto
	LastContactFieldName = "lastContact"
	// AssetPrefix prefix for assetUrn
	AssetPrefix = "psc:device"
	// UrnSeparator urnSeparator
	UrnSeparator = ":"
)

// LastContactPublisher publishes to elastic search.
type LastContactPublisher struct {
	Log                  *zerolog.Logger
	esBulkProcessorsV5   *ElasticSearchBulkProcessorsV5
	esBulkProcessorsV6   *ElasticSearchBulkProcessorsV6
	statsReporter        reporting.WavefrontMetricsReporter
	conf                 *config.Configurations
	assetRegistryEnabled bool
	kinesisPublisher     *KinesisPublisher
}

// NewLCTPublisher creates a new LastContactTime publisher. This will be created per shard.
func NewLCTPublisher(zlog *zerolog.Logger, esBulkProcessorsV5 *ElasticSearchBulkProcessorsV5, esBulkProcessorsV6 *ElasticSearchBulkProcessorsV6, statsReporter reporting.WavefrontMetricsReporter, conf *config.Configurations, pushToAR bool, kinesisPublisher *KinesisPublisher) LastContactPublisher {
	s := LastContactPublisher{
		Log:                  zlog,
		esBulkProcessorsV5:   esBulkProcessorsV5,
		esBulkProcessorsV6:   esBulkProcessorsV6,
		statsReporter:        statsReporter,
		conf:                 conf,
		assetRegistryEnabled: pushToAR,
		kinesisPublisher:     kinesisPublisher,
	}
	return s
}

// Publish -- Takes the records received from last contact stream and publishes them onto elasticsearch
// (and the asset registry stream, if enabled)
func (p *LastContactPublisher) Publish(batches <-chan *LctBatch) {
	for b := range batches {
		var assetDataChangeEvents []*assetupdater.AssetDataChangeEvent
		lctEventCountPerEnv := make(map[string]int64)

		p.Log.Debug().Msgf("Publish - DeviceDataList: %d", len(b.DeviceDataList))
		for k, lct := range b.DeviceDataList {
			orgKey, err := orgkey.Decode(lct.OrgKey)
			if err != nil {
				p.Log.Error().Msgf("Failed to decode orgKey %v", lct.OrgKey)
			}
			if orgKey == nil {
				p.Log.Error().Msg("Unexpected, orgKey is nil")
				continue
			}

			if p.assetRegistryEnabled {
				assetDataChangeEvents = append(assetDataChangeEvents, p.prepareAssetDataChaneEvent(lct))

				// Used for WF Metrics collection
				env := orgKey.Env()
				if count, ok := lctEventCountPerEnv[env]; ok {
					lctEventCountPerEnv[env] = count + 1
				} else {
					lctEventCountPerEnv[env] = 1
				}
			}

			// add entry to elasticsearch for bulk update
			p.publishAsBulkUpdate(k == len(b.DeviceDataList)-1, orgKey, lct.DeviceId, lct.LastContactTimestamp)
		}

		if p.assetRegistryEnabled {
			// push records in this batch to asset registry
			p.kinesisPublisher.PutRecords(assetDataChangeEvents, lctEventCountPerEnv)
		}
	}
}

func (p *LastContactPublisher) publishAsBulkUpdate(logThis bool, orgKey *orgkey.OrgKey, deviceID, lastContactTimestamp int64) {
	orgIDStr := strconv.Itoa(orgKey.OrgID())
	esVersion := p.conf.EsVersionEnvMap[orgKey.Env()]

	stub := DeviceStatsStub{
		OrganizationID: int64(orgKey.OrgID()),
		DeviceID:       deviceID,
		LastContact:    lastContactTimestamp,

		// From https://gitlab.bit9.local/cb-defense/cbd/-/blob/main/apps/deviceServices/src/main/java/com/scargo/app/services/hello/StreamedDeviceLastContactTracker.java#L52-54
		// and https://gitlab.bit9.local/cb-defense/cbd/-/blob/main/apps/deviceServices/src/main/java/com/scargo/app/common/DeviceLastContactTracker.java#L185-187
		// since we have received a heartbeat, reset deregisterAfterTime
		// this is required for the CWP workloads
		DeregisterAfterTime: 0,
	}

	switch esVersion {
	case "es_6":
		bulkRequest := elasticV6.NewBulkUpdateRequest()
		bulkRequest.Type("device").
			Id(fmt.Sprintf("%s-%s", orgIDStr, strconv.FormatInt(deviceID, 10))).
			Routing(orgIDStr).DetectNoop(false).
			Index(indexName()).Doc(&stub)
		processor := p.esBulkProcessorsV6.getOrCreateProcessor(orgKey.Env(), p.afterBulkUpdateV6(orgKey.Env()))
		p.Log.Info().Msgf("Push via ES6: %+v", bulkRequest)
		p.Log.Debug().Msgf("   Stats: %+v", processor.Stats())
		if logThis {
			p.Log.Info().Msgf("Stats[Succeeded=%v, Committed=%v, Failed=%v]", processor.Stats().Succeeded, processor.Stats().Committed, processor.Stats().Failed)
		}
		processor.Add(bulkRequest)

	default:
		bulkRequest := elasticV5.NewBulkUpdateRequest()
		bulkRequest.Type("device").
			Id(fmt.Sprintf("%s-%s", orgIDStr, strconv.FormatInt(deviceID, 10))).
			Routing(orgIDStr).VersionType("force").
			Version(1).DetectNoop(false).
			Index(indexName()).Doc(&stub)
		processor := p.esBulkProcessorsV5.getOrCreateProcessor(orgKey.Env(), p.afterBulkUpdateV5(orgKey.Env()))
		p.Log.Info().Msgf("Push via ES5: %+v", bulkRequest)
		p.Log.Debug().Msgf("   Stats: %+v", processor.Stats())
		if logThis {
			p.Log.Info().Msgf("Stats[Succeeded=%v, Committed=%v, Failed=%v]", processor.Stats().Succeeded, processor.Stats().Committed, processor.Stats().Failed)
		}
		processor.Add(bulkRequest)
	}
}

func (p *LastContactPublisher) prepareAssetDataChaneEvent(lct *devicelastcontacttime.DeviceLastContactTime) *assetupdater.AssetDataChangeEvent {
	assetUrn := AssetPrefix + UrnSeparator + lct.OrgKey + UrnSeparator + strconv.FormatInt(lct.DeviceId, 10)
	p.Log.Debug().Msgf("Preparing AssetDataChangeEvent for asset_urn = %v", assetUrn)
	uuid, err := securerandom.Uuid()
	if err != nil {
		p.Log.Fatal().Msgf("Error in generating UUID for publishing LCT event to AR stream - %v", err)
	}

	lastContactField := make(map[string]*assetupdater.AssetDataChangeEvent_Field)
	lastContactField[LastContactFieldName] = &assetupdater.AssetDataChangeEvent_Field{
		ChangeType: assetupdater.AssetDataChangeEvent_SET_VALUE,
		Values:     []string{strconv.FormatInt(lct.LastContactTimestamp, 10)},
	}
	return &assetupdater.AssetDataChangeEvent{
		Id:                   uuid,
		AssetUrn:             assetUrn,
		PublisherId:          p.conf.PublisherID,
		EventCreateTimestamp: timestamppb.Now(),
		Type:                 assetupdater.AssetDataChangeEvent_UPDATE,
		Fields:               lastContactField,
	}
}

func (p *LastContactPublisher) afterBulkUpdateV5(env string) elasticV5.BulkAfterFunc {
	return func(env string) elasticV5.BulkAfterFunc {
		return func(executionId int64, requests []elasticV5.BulkableRequest, response *elasticV5.BulkResponse, err error) {
			tags := make(map[string]string)
			tags["env"] = env
			if env == "local" {
				tags["env"] = "dev01"
			}

			if response != nil {
				p.Log.Info().Msgf("Time taken = %v", response.Took)
				if response.Errors {
					// es_bulk_update.failedRecords - This metric represents number of failure records in a batch
					stats.UpdateMeter(p.statsReporter, "es_bulk_update.failedRecords", tags, int64(len(response.Failed())))
					p.Log.Error().Msgf("%d errors seen during ES v5 bulk update:", len(response.Failed()))
					for _, value := range response.Failed() {
						p.Log.Error().Msgf("err %+v", value.Error)
					}
				}

				// es_bulk_update.successRecords - This metric represents number of succeeded records in a batch
				stats.UpdateMeter(p.statsReporter, "es_bulk_update.successRecords", tags, int64(len(response.Succeeded())))
				stats.UpdateTimer(p.statsReporter, "es_bulk_update.time", tags, time.Duration(response.Took)*time.Millisecond)
			} else if err != nil {
				p.Log.Error().Msgf("Bulk update response was nil: %v", err)
				stats.UpdateMeter(p.statsReporter, "es_bulk_update.error", tags, 1)
				stats.UpdateMeter(p.statsReporter, "es_bulk_update.failedRecords", tags, int64(p.conf.EsBatchSizeEnvMap[env]))
			}
		}
	}(env)
}

func (p *LastContactPublisher) afterBulkUpdateV6(env string) elasticV6.BulkAfterFunc {
	return func(env string) elasticV6.BulkAfterFunc {
		return func(executionId int64, requests []elasticV6.BulkableRequest, response *elasticV6.BulkResponse, err error) {
			tags := make(map[string]string)
			tags["env"] = env
			if env == "local" {
				tags["env"] = "dev01"
			}

			if response != nil {
				p.Log.Info().Msgf("Time taken = %v", response.Took)
				if response.Errors {
					stats.UpdateMeter(p.statsReporter, "es_bulk_update.failedRecords", tags, int64(len(response.Failed())))
					p.Log.Error().Msgf("%d errors seen during ES v6 bulk update:", len(response.Failed()))
					for i, value := range response.Failed() {
						p.Log.Error().Msgf("Error #%d: %+v", i, value.Error)
					}
				}
				stats.UpdateMeter(p.statsReporter, "es_bulk_update.successRecords", tags, int64(len(response.Succeeded())))
				stats.UpdateTimer(p.statsReporter, "es_bulk_update.time", tags, time.Duration(response.Took)*time.Millisecond)
			} else if err != nil {
				p.Log.Error().Msgf("Bulk update response was nil: %v", err)
				stats.UpdateMeter(p.statsReporter, "es_bulk_update.error", tags, 1)
				stats.UpdateMeter(p.statsReporter, "es_bulk_update.failedRecords", tags, int64(p.conf.EsBatchSizeEnvMap[env]))
			}
		}
	}(env)
}

func indexName() string {
	return "a_devicestats"
}

// DeviceStatsStub stub to update lastContact info for device
type DeviceStatsStub struct {
	OrganizationID      int64 `json:"organizationId"`
	DeviceID            int64 `json:"deviceId"`
	LastContact         int64 `json:"lastContact"`
	DeregisterAfterTime int64 `json:"deregisterAfterTime"`
}
